/****************************************************************************
**
** SqlSync - Comparing and synchronizing data in SQL databases
** Copyright (c) 2006-2008, SILVERCODERS(R)
** http://silvercoders.com
**
** Project homepage: http://silvercoders.com/index.php?page=sqlsync
** Project authors:  Pawel 'pawel_p' Pluciennik
**
** This program may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file COPYING.GPL included in the
** packaging of this file.
**
** Please remember that any attempt to workaround the GNU General Public
** License using wrappers, pipes, client/server protocols, and so on
** is considered as license violation. If your program, published on license
** other than GNU General Public License version 2, calls some part of this
** code directly or indirectly, you have to buy commercial license.
** If you do not like our point of view, simply do not use the product.
**
** Licensees holding valid commercial license for this product
** may use this file in accordance with the license published by
** SILVERCODERS and appearing in the file COPYING.COM
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*****************************************************************************/

#include "sqlsync.h"

#include "sqlsync_internal.h"

#include "sqlsync_debug.h"

#include "sqlsync_translation.h"
#include "sqlsync_compare_records.h"

#include "sqlsync_user_internal.h"

#include "server/sqlsync_sqlserver.h"
#include "server/sqlsync_sqlserver_helpers.h"
#include "server/sqlsync_dbstruct.h"
#include "server/sqlsync_query.h"
#include "sqlsync_synchronization.h"

sqlsync_info_ptr
sqlsync_init (SQLSYNC_RETURN * error)
{

  *error = SQLSYNC_SUCCESS;

  sqlsync_info *info = std_malloc (sizeof (sqlsync_info));


#ifdef DEBUG
  sqlsync_debug ("sqlsync_init completed\n");
#endif

  info->run_queries = SQLSYNC_FALSE;

  info->server[SERVER_SRC] = NULL;
  info->server[SERVER_DST] = NULL;
  info->cs_ignore = SQLSYNC_FALSE;


  c_create_vtbls ();
  sqlsync_translation_init ();
  sqlsync_send_user_notification_init ();
  sqlsync_user_notification_register (sqlsync_send_user_notification_synchronization_handler, info);

  return info;
}

SQLSYNC_RETURN
sqlsync_destroy (sqlsync_info_ptr info_ptr)
{
  SQLSYNC_RETURN ret = SQLSYNC_SUCCESS;
  sqlsync_info *info = (sqlsync_info *) info_ptr;

  if (info->server[SERVER_SRC] != NULL)
    ret |= sqlsync_sqlserver_destroy (info->server[SERVER_SRC]);

  if (info->server[SERVER_DST] != NULL)
    ret |= sqlsync_sqlserver_destroy (info->server[SERVER_DST]);


  if (ret != SQLSYNC_SUCCESS)
  {
#ifdef DEBUG
    sqlsync_debug ("sqlsync_destroy failed (ret %x)\n", ret);
#endif

    sqlsync_send_error_notification2 (_tr ("Internal error in sqlsync_destroy"));

    return SQLSYNC_FAILED_DESTROY;
  }

#ifdef DEBUG
  sqlsync_debug ("sqlsync_destroy completed\n");
#endif

  std_free (info);

  sqlsync_send_user_notification_free ();
  c_destroy_vtbls ();

  return SQLSYNC_SUCCESS;
}

void sqlsync_ignore_case_sensitive(sqlsync_info_ptr info_ptr, SQLSYNC_RETURN ignore)
{
  sqlsync_info *info = (sqlsync_info *) info_ptr;

  info->cs_ignore = ignore;
}

SQLSYNC_RETURN
sqlsync_connect (sqlsync_info_ptr info_ptr, sqlsync_server nr_server,
                 c_string_elem_ptr dsn, c_string_elem_ptr user, c_string_elem_ptr password)
{
  SQLSYNC_RETURN ret;
  sqlsync_info *info = (sqlsync_info *) info_ptr;


  ret = sqlsync_sqlserver_create_connect (&info->server[nr_server], dsn, user, password, info->cs_ignore);
  if (ret != SQLSYNC_SUCCESS)
    return ret;

#ifdef DEBUG
  sqlsync_debug ("Server type %ld\n", sqlsync_sqlserver_type (info->server[nr_server]));
  sqlsync_debug ("sqlsync_connect completed (", 0);
  sqlsync_prints (SQLSYNC_DEBUG, dsn);
  sqlsync_printf (SQLSYNC_DEBUG, " is ", 0);
  sqlsync_printf (SQLSYNC_DEBUG, ")\n", 0);
#endif
  return SQLSYNC_SUCCESS;
}

c_cmp_ptr
sqlsync_compare_tables (sqlsync_info_ptr info_ptr, sqlsync_table_columns * tc_src, sqlsync_table_columns * tc_dst)
{
  c_cmp_ptr result = NULL;

  sqlsync_info *info = (sqlsync_info *) info_ptr;

  c_grid_data_ptr grid_src = c_grid_data_init (10);
  c_grid_data_ptr grid_dst = c_grid_data_init (10);


  /* Initializing result informations */
  result = c_cmp_init (sqlsync_sqlserver_type (info->server[SERVER_DST]));

  /* Storing columns */
  (void) c_cmp_add_columns (result, tc_dst->columns);
  /* Storing table name */
  (void) c_cmp_set_name (result, tc_dst->table);

#ifdef DEBUG
  sqlsync_debug ("Writing table name %s" "to c_cmp_ptr results[*size] %ld \n", tc_dst->table->string, 0);
#endif

  sqlsync_send_string_notification (SQLSYNC_USER_COMPARE_TABLES_BEGIN, tc_src->table);


  (void) sqlsync_find_set_primary_keys (info->server[SERVER_SRC], info->server[SERVER_DST],
                                        result->table_name, result->columns, &result->primary_keys,
                                        &result->primary_keys_pos);

  (void) sqlsync_run_SELECT_queries (info, tc_src, tc_dst);

  (void) sqlsync_get_rows (info->server[SERVER_SRC], info->server[SERVER_DST], tc_src->columns,
                           tc_dst->columns,
                           grid_src, grid_dst, result->primary_keys_pos, &result->equals);


  (void) sqlsync_compare_grids (grid_src, grid_dst, result->primary_keys_pos,
                                result->row_miss, result->row_add,
                                result->row_diff_src, result->row_diff_dst, &result->equals);

  sqlsync_send_string_notification (SQLSYNC_USER_COMPARE_TABLES_END, tc_src->table);


  (void) c_grid_data_destroy (grid_dst);
  (void) c_grid_data_destroy (grid_src);

  return result;
}

void
sqlsync_print_error (sqlsync_info_ptr info_ptr, sqlsync_server nr_server)
{
  sqlsync_sqlserver_print_error (((sqlsync_info *) info_ptr)->server[nr_server]);
}


uint32
sqlsync_compare_databases (sqlsync_info_ptr info_ptr, c_vector_strings_ptr tables,
                           c_grid_columns_ptr tables_columns,
                           SQLSYNC_RETURN show_summary,
                           SQLSYNC_RETURN show_queries, SQLSYNC_RETURN run_queries)
{
  uint32 nr_tables = 0;
  uint32 i = 0;
  uint32 size = 0;

  sqlsync_tables_columns valid_src;
  sqlsync_tables_columns valid_dst;

  c_cmp_ptr result = NULL;

  sqlsync_info *info = (sqlsync_info *) info_ptr;


  (void) sqlsync_retrieve_valid_tables_columns (info->server[SERVER_SRC], info->server[SERVER_DST],
                                                tables,
                                                tables_columns,
                                                &valid_src, &valid_dst);

  nr_tables = c_vector_strings_size (valid_src.valid_tables);


  for (i = 0; i < nr_tables; i++)
  {
    sqlsync_table_columns v_src;
    sqlsync_table_columns v_dst;
    v_src.table = c_vector_strings_pos (valid_src.valid_tables, i);
    v_src.columns = c_grid_columns_pos (valid_src.valid_tables_columns, i);

    v_dst.table = c_vector_strings_pos (valid_dst.valid_tables, i);
    v_dst.columns = c_grid_columns_pos (valid_dst.valid_tables_columns, i);

    if ((v_dst.table == NULL || v_dst.columns == NULL) ||
        (v_src.table == NULL || v_src.columns == NULL))
      continue;

    /* Create results table (additional, missing, different rows) 
     * for specified tables
     */
    result = sqlsync_compare_tables (info, &v_src, &v_dst);
    if (result != NULL)
    {
      /* Increase numer of  results */
      size++;
    }

    if (show_summary == SQLSYNC_TRUE)
    {
      sqlsync_send_table_summary (result);
    }

    if (run_queries == SQLSYNC_TRUE)
    {
      sqlsync_create_run_synchronization_queries (info_ptr, result);
    }

    if (show_queries == SQLSYNC_TRUE)
    {
      sqlsync_create_synchronization_queries (result, SQLSYNC_FALSE);
    }


#ifdef DEBUG
    sqlsync_debug ("Increased number of results to %ld\n", size);
#endif

    c_cmp_destroy (&result);
    result = NULL;
  }

  c_vector_strings_destroy (valid_src.valid_tables);
  c_grid_columns_destroy (valid_src.valid_tables_columns);

  c_vector_strings_destroy (valid_dst.valid_tables);
  c_grid_columns_destroy (valid_dst.valid_tables_columns);


#ifdef DEBUG
  sqlsync_debug ("sqlsync_compare_databases completed\n");
#endif

  return size;
}

SQLSYNC_RETURN
sqlsync_run_query_file (sqlsync_info_ptr /*@null@ */ info_ptr,
                        sqlsync_server nr_server, c_string_elem_ptr name)
{
  return sqlsync_execute_query_file (((sqlsync_info *) info_ptr)->server[nr_server], name);
}


void
sqlsync_parse_tables_columns (c_string_elem_ptr tables_columns,
                              c_vector_strings_ptr * tables, c_grid_columns_ptr * columns)
{
  char c = '\0';
  uint32 i = 0;
  char *table_begin = NULL;
  uint32 table_size = 0;

  char *table_next_begin = NULL;
  uint32 table_next_size = 0;

  char *column_begin = NULL;
  uint32 column_size = 0;

  char *column_next_begin = NULL;
  uint32 column_next_size = 0;

  short table_found = 0;
  short column_found = 0;
  short columns_read = 0;
  c_vector_columns_ptr table_columns = NULL;


  if (tables_columns == NULL || tables_columns->string == NULL)
  {
    *tables = NULL;
    *columns = NULL;
    return;
  }

  sqlsync_send_info_notification2 (_tr ("Parsing tables and columns passed from command line"));

  *tables = c_vector_strings_init (10);
  *columns = c_grid_columns_init (10);

  table_begin = tables_columns->string;

#ifdef DEBUG
  sqlsync_debug ("Tables and columns string [", 0);
  sqlsync_prints (SQLSYNC_DEBUG, tables_columns);
  sqlsync_printf (SQLSYNC_DEBUG, "]\n", 0);
#endif

  for (i = 0; i < tables_columns->length; i++)
  {
    c = tables_columns->string[i];
    switch (c)
    {
      case '[':
      {
        column_begin = tables_columns->string + i + 1;
        column_size = 0;
        column_found = 1;

        if (table_begin != NULL)
        {
          table_size = tables_columns->string + i - table_begin;
          table_next_begin = NULL;
          table_found = 1;
        }

        else
        {
#ifdef DEBUG
          sqlsync_debug ("Error in parsed string '[' found,"
                         " but table name wasn't specified before\n");
#endif
        }
        break;
      }
      case ']':
      {
        if (column_begin != NULL)
        {
          column_size = tables_columns->string + i - column_begin;
          column_next_begin = NULL;
          column_next_size = 0;
          columns_read = 1;
          column_found = 1;
        }

        else
        {
#ifdef DEBUG
          sqlsync_debug ("Error in parsed string, '[' not found\n");
#endif
        }
        break;
      }
      case ',':
      {
        if (table_begin != NULL && table_size == 0)
        {
          table_size = tables_columns->string + i - table_begin;
          table_next_begin = tables_columns->string + i + 1;
          table_found = 1;
          columns_read = 1;

        }
        else if (column_begin != NULL)
        {
          column_size = tables_columns->string + i - column_begin;
          column_next_begin = tables_columns->string + i + 1;
          column_found = 1;
        }
        else
          table_begin = tables_columns->string + i + 1;
        break;
      }
      default:
      {
        if (table_begin != NULL && i == tables_columns->length - 1)
        {
          table_size = tables_columns->string + i - table_begin + 1;
          table_found = 1;
          column_found = 0;
          columns_read = 1;
        }
        break;
      }
    }


    if (column_begin != NULL && column_size != 0 && (columns_read == 1 || column_found == 1))
    {
      c_column_elem_ptr elem;
      c_string_elem column;

      column_found = 0;
      column.string = column_begin;
      column.length = column_size;
      column.reserved_length = column_size;
      column._type = c_string_elem_type;

      elem = c_column_elem_create (&column, 0, 0, 0);
#ifdef DEBUG
      sqlsync_debug ("Found (%s) column\n", elem->column);
#endif
      (void) c_vector_columns_add_last (table_columns, elem);


      column_begin = column_next_begin;
      column_size = column_next_size;

    }

    if (table_begin != NULL && table_size != 0 && table_found == 1)
    {
      table_found = 0;

      if (table_columns == NULL)
      {
        table_columns = c_vector_columns_init (10);
      }
    }

    if (table_begin != NULL && table_size != 0 && columns_read == 1)
    {
      c_string_elem table;

#ifdef DEBUG
      char *tmp;

      tmp = std_malloc (table_size + 1);
      strncpy (tmp, table_begin, table_size);
      tmp[table_size] = '\0';

      sqlsync_debug ("Found (%s) table\n", tmp);
      std_free (tmp);
#endif

      table.string = table_begin;
      table.length = table_size;
      table._type = c_string_elem_type;
      table.reserved_length = table_size;

      (void) c_vector_strings_add_last_copy (*tables, &table);

      table_begin = table_next_begin;
      table_size = table_next_size;
      column_begin = NULL;
      column_size = 0;
      columns_read = 0;
      table_found = 0;
      column_found = 0;

#ifdef DEBUG
      sqlsync_debug ("Specified %d columns\n", c_vector_columns_size (table_columns));
#endif

      (void) c_grid_columns_add_last_copy (*columns, table_columns);

      (void) c_vector_columns_destroy (table_columns);
      table_columns = NULL;

#ifdef DEBUG
      sqlsync_debug ("Number of columns %ld\n",
                     c_vector_data_size (c_grid_columns_pos
                                         (*columns, c_grid_columns_size (*columns) - 1)));

      sqlsync_debug ("(sqlsync) --------------\n");
#endif
    }

  }

  sqlsync_send_info_notification2 (_tr ("Parsing tables and columns completed"));

}
